from __future__ import annotations
from typing import Callable, TYPE_CHECKING
from qtpy import QtWidgets as QtW, QtGui, QtCore
from qtpy.QtWidgets import QAction
from qtpy.QtCore import Qt, Signal

from ._utils import create_temporal_line_edit

from .._utils import search_name_from_qmenu

if TYPE_CHECKING:
    from .._table import QBaseTable


class QTabbedTableStack(QtW.QTabWidget):
    """Tab widget used for table stack."""

    currentTableChanged = Signal(int)  # index
    itemDropped = Signal(object)  # dropped item info
    itemMoved = Signal(int, int)  # source index, destination index
    tableRenamed = Signal(int, str)  # index
    tableRemoved = Signal(int)  # index
    tablePassed = Signal(object, int, object)  # source widget, tab_id, target widget

    def __init__(
        self,
        parent=None,
        tab_position: str = "top",
    ):
        super().__init__(parent)
        if tab_position == "top":
            pass
        elif tab_position == "left":
            from ._sidebar import QLeftSideBar

            self.setTabBar(QLeftSideBar(self))
            self.setTabPosition(QtW.QTabWidget.TabPosition.West)
        elif tab_position == "bottom":
            self.setTabPosition(QtW.QTabWidget.TabPosition.South)
        elif tab_position == "right":
            from ._sidebar import QRightSideBar

            self.setTabBar(QRightSideBar(self))
            self.setTabPosition(QtW.QTabWidget.TabPosition.East)
        else:
            raise ValueError(f"Unknown position {tab_position!r}.")

        self.setAcceptDrops(True)
        self.setMovable(True)
        self.tabBar().setMouseTracking(True)
        self.setStyleSheet("QTabWidget::pane { border: 1px solid #C4C4C3; top: -1px; }")

        self.tabBar().setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.tabBar().customContextMenuRequested.connect(self.showContextMenu)
        self.currentChanged.connect(self.currentTableChanged.emit)
        self.tabCloseRequested.connect(self.takeTable)
        self.tabCloseRequested.connect(self.tableRemoved.emit)
        # NOTE: arguments are not (from, to). Bug in Qt??
        self.tabBar().tabMoved.connect(lambda a, b: self.itemMoved.emit(b, a))
        self.tabBarDoubleClicked.connect(self.enterEditingMode)
        self.installContextMenu()

        self._line: QtW.QLineEdit | None = None  # temporal QLineEdit for editing tabs

    def addTable(self, table: QBaseTable, name: str = "None"):
        """Add `table` to stack as name `name`."""
        self.addTab(table, name)
        self.tabBar().setTabData(self.count() - 1, table)
        return None

    def takeTable(self, index: int) -> QBaseTable:
        """Remove table at `index` and return it."""
        return self.removeTab(index)

    def renameTable(self, index: int, name: str):
        """Rename table at `index` to `name`."""
        return self.setTabText(index, name)

    def tableIndex(self, table: QBaseTable) -> int:
        """Get the index of `table`."""
        for i in range(self.count()):
            data = self.tabBar().tabData(i)
            if data is table:
                break
        else:
            raise ValueError(f"Table {table!r} not found.")
        return i

    def tableAtIndex(self, i: int) -> QBaseTable:
        """Get the table at `i`."""
        return self.tabBar().tabData(i)

    def tableAt(self, pos: QtCore.QPoint) -> QBaseTable | None:
        """Return table at position."""
        index = self.tabBar().tabAt(pos)
        # TODO: bug when position is east or south
        if index == -1:
            return None
        return self.tableAtIndex(index)

    def moveTable(self, src: int, dst: int):
        """Move table from `src` to `dst`."""
        return self.tabBar().moveTab(src, dst)

    def dragEnterEvent(self, e: QtGui.QDragEnterEvent) -> None:
        # This override is necessary for accepting drops from files.
        e.accept()

        if e.source().parentWidget() is not self:
            return

        self._entering_tab_index = self.indexOf(self.widget(self._moving_tab_index))

    def dropEvent(self, e: QtGui.QDropEvent) -> None:
        mime = e.mimeData()
        text = mime.text()
        if text:
            self.itemDropped.emit(text)

        source_widget: QTabbedTableStack = e.source().parentWidget()
        tab_id = source_widget._entering_tab_index
        if source_widget is self:
            return super().dropEvent(e)

        e.setDropAction(Qt.DropAction.MoveAction)
        e.accept()

        self.tablePassed.emit(source_widget, tab_id, self)
        return super().dropEvent(e)

    def mouseMoveEvent(self, e: QtGui.QMouseEvent):
        pos_global = self.mapToGlobal(e.pos())
        tabbar = self.tabBar()
        pos_intab = tabbar.mapFromGlobal(pos_global)
        self._moving_tab_index = tabbar.tabAt(e.pos())
        if self._moving_tab_index < 0:
            return

        tabrect = self.tabRect(self._moving_tab_index)

        pixmap = QtGui.QPixmap(tabrect.size())
        tabbar.render(pixmap, QtCore.QPoint(), QtGui.QRegion(tabrect))
        mime = QtCore.QMimeData()
        drag = QtGui.QDrag(tabbar)
        drag.setMimeData(mime)
        drag.setPixmap(pixmap)
        cursor = QtGui.QCursor(Qt.CursorShape.OpenHandCursor)
        drag.setHotSpot(e.pos() - pos_intab)
        drag.setDragCursor(cursor.pixmap(), Qt.DropAction.MoveAction)
        drag.exec_(Qt.DropAction.MoveAction)

    def mousePressEvent(self, e: QtGui.QMouseEvent) -> None:
        if self._line is not None:
            self._line.setHidden(True)
        return super().mousePressEvent(e)

    def mouseDoubleClickEvent(self, e: QtGui.QMouseEvent) -> None:
        if e.button() != Qt.MouseButton.LeftButton:
            return
        return super().mouseDoubleClickEvent(e)

    def dragLeaveEvent(self, e: QtGui.QDragLeaveEvent):
        e.accept()
        return None

    def tabRect(self, index: int):
        """Get QRect of the tab at index."""
        rect = self.tabBar().tabRect(index)

        # NOTE: East/South tab returns wrong value (Bug in Qt?)
        if self.tabPosition() == QtW.QTabWidget.TabPosition.East:
            w = self.rect().width() - rect.width()
            rect.translate(w, 0)
        elif self.tabPosition() == QtW.QTabWidget.TabPosition.South:
            h = self.rect().height() - rect.height()
            rect.translate(0, h)

        return rect

    def enterEditingMode(self, index: int):
        """Enter edit table name mode."""
        rect = self.tabRect(index)

        # set geometry
        line = create_temporal_line_edit(rect, self, self.tabText(index))
        self._line = line

        @self._line.editingFinished.connect
        def _():
            self._line.setHidden(True)
            text = self._line.text()
            self.setTabText(index, text)
            self.tableRenamed.emit(index, text)

        return None

    def installContextMenu(self):
        """Install the default contextmenu."""
        self._qt_context_menu = QTabContextMenu(self)

        @self.registerAction("Copy all")
        def _copy(index: int):
            table = self.tableAtIndex(index)
            h, w = table.tableShape()
            table.setSelections([(slice(0, h), slice(0, w))])
            table.copyToClipboard(headers=True)
            table.setSelections([])

        self.registerAction("Rename")(self.enterEditingMode)

        @self.registerAction("Delete")
        def _delete(index: int):
            self.takeTable(index)
            self.tableRemoved.emit(index)

    def registerAction(self, location: str):
        locs = location.split(">")
        menu = self._qt_context_menu
        for loc in locs[:-1]:
            a = search_name_from_qmenu(menu, loc)
            if a is None:
                menu = menu.addMenu(loc)
            else:
                menu = a.menu()
                if menu is None:
                    i = locs.index(loc)
                    err_loc = ">".join(locs[:i])
                    raise TypeError(f"{err_loc} is not a menu.")

        def wrapper(f: Callable):
            action = QAction(locs[-1], self)
            action.triggered.connect(lambda: f(self._qt_context_menu._current_index))
            menu.addAction(action)
            return f

        return wrapper

    def showContextMenu(self, pos: QtCore.QPoint) -> None:
        """Execute contextmenu."""
        table = self.tableAt(pos)
        if table is None:
            return
        index = self.tableIndex(table)
        self._qt_context_menu.execAtIndex(QtGui.QCursor().pos(), index)
        return


class QTabContextMenu(QtW.QMenu):
    def __init__(self, parent=None):
        super().__init__(parent)
        self._current_index = None

    def execAtIndex(self, pos: QtCore.QPoint, index: int):
        """Execute contextmenu at index."""
        self._current_index = index
        try:
            self.exec_(pos)
        finally:
            self._current_index = None
        return None



from __future__ import annotations
from typing import Callable
from qtpy.QtWidgets import QMessageBox, QAction
from qtpy import QtWidgets as QtW
from enum import Enum


class MessageBoxMode(Enum):
    ERROR = "error"
    WARNING = "warn"
    INFO = "info"
    QUESTION = "question"
    ABOUT = "about"


_QMESSAGE_MODES = {
    MessageBoxMode.ERROR: QMessageBox.critical,
    MessageBoxMode.WARNING: QMessageBox.warning,
    MessageBoxMode.INFO: QMessageBox.information,
    MessageBoxMode.QUESTION: QMessageBox.question,
    MessageBoxMode.ABOUT: QMessageBox.about,
}


def show_messagebox(
    mode: str | MessageBoxMode = MessageBoxMode.INFO,
    title: str = None,
    text: str = None,
    parent=None,
) -> bool:
    """
    Freeze the GUI and open a messagebox dialog.

    Parameters
    ----------
    mode : str or MessageBoxMode, default is MessageBoxMode.INFO
        Mode of message box. Must be "error", "warn", "info", "question" or "about".
    title : str, optional
        Title of messagebox.
    text : str, optional
        Text in messagebox.
    parent : QWidget, optional
        Parent widget.

    Returns
    -------
    bool
        If "OK" or "Yes" is clicked, return True. Otherwise return False.
    """
    show_dialog = _QMESSAGE_MODES[MessageBoxMode(mode)]
    result = show_dialog(parent, title, text)
    return result in (QMessageBox.Ok, QMessageBox.Yes)

def to_action(f: Callable, parent=None) -> QAction:
    action = QAction(f.__name__.replace("_", " "), parent)
    action.triggered.connect(f)
    return action

def search_name_from_qmenu(qmenu: QtW.QMenu | QtW.QMenuBar, name: str):
    for a in qmenu.children():
        if isinstance(a, QAction) and a.text() == name:
            return a
    return None



from __future__ import annotations
from typing import Any, overload
from qtpy import QtWidgets as QtW, QtGui, QtCore
from qtpy.QtCore import Signal, Qt
import pandas as pd

import numpy as np
from ._model_base import AbstractDataFrameModel
from ..._utils import show_messagebox
from ....types import FilterType, ItemInfo, Selections

# Flags
_EDITABLE = (
    QtW.QAbstractItemView.EditTrigger.EditKeyPressed
    | QtW.QAbstractItemView.EditTrigger.DoubleClicked
)
_READ_ONLY = QtW.QAbstractItemView.EditTrigger.NoEditTriggers
_SCROLL_PER_PIXEL = QtW.QAbstractItemView.ScrollMode.ScrollPerPixel


class _QTableViewEnhanced(QtW.QTableView):
    selectionChangedSignal = Signal()
    rightClickedSignal = Signal(QtCore.QPoint)

    def __init__(self, parent=None):
        super().__init__(parent)
        self._last_pos: QtCore.QPoint | None = None
        self._was_right_dragging: bool = False
        self._zoom = 1.0
        self._initial_font_size = self.font().pointSize()
        vheader, hheader = self.verticalHeader(), self.horizontalHeader()
        self.setFrameStyle(QtW.QFrame.Shape.Box)
        vheader.setFrameStyle(QtW.QFrame.Shape.Box)
        hheader.setFrameStyle(QtW.QFrame.Shape.Box)
        vheader.resize(16, vheader.height())
        self.setStyleSheet("QHeaderView::section { border: 1px solid black}")
        vheader.setMinimumSectionSize(0)
        hheader.setMinimumSectionSize(0)
        vheader.font().setPointSize(self._initial_font_size)
        hheader.font().setPointSize(self._initial_font_size)

        vheader.setDefaultSectionSize(28)
        hheader.setDefaultSectionSize(100)

        hheader.setSectionResizeMode(QtW.QHeaderView.ResizeMode.Fixed)
        vheader.setSectionResizeMode(QtW.QHeaderView.ResizeMode.Fixed)

        self._initial_section_size = (
            hheader.defaultSectionSize(),
            vheader.defaultSectionSize(),
        )
        self.setZoom(1.0)  # initialize

    def selectionChanged(
        self,
        selected: QtCore.QItemSelection,
        deselected: QtCore.QItemSelection,
    ) -> None:
        """Evoked when table selection range is changed."""
        self.selectionChangedSignal.emit()
        return super().selectionChanged(selected, deselected)

    def mousePressEvent(self, e: QtGui.QMouseEvent) -> None:
        """Register clicked position"""
        if e.button() == Qt.MouseButton.RightButton:
            self._last_pos = e.pos()
            self._was_right_dragging = False
        return super().mousePressEvent(e)

    def mouseMoveEvent(self, e: QtGui.QMouseEvent) -> None:
        """Scroll table plane when mouse is moved with right click."""
        if self._last_pos is not None:
            pos = e.pos()
            dy = pos.y() - self._last_pos.y()
            dx = pos.x() - self._last_pos.x()
            self.verticalScrollBar().setValue(self.verticalScrollBar().value() - dy)
            self.horizontalScrollBar().setValue(self.horizontalScrollBar().value() - dx)
            self._last_pos = pos
            self._was_right_dragging = True
        return super().mouseMoveEvent(e)

    def mouseReleaseEvent(self, e: QtGui.QMouseEvent) -> None:
        """Delete last position."""
        if e.button() == Qt.MouseButton.RightButton and not self._was_right_dragging:
            self.rightClickedSignal.emit(e.pos())
        self._last_pos = None
        self._was_right_dragging = False
        return super().mouseReleaseEvent(e)

    def keyPressEvent(self, e: QtGui.QKeyEvent) -> None:
        """Evoke parent keyPressEvent."""
        parent = self.parent()
        if isinstance(parent, QBaseTable):
            parent.keyPressEvent(e)
        return super().keyPressEvent(e)

    def zoom(self) -> float:
        """Get current zoom factor."""
        return self._zoom

    def setZoom(self, value: float) -> None:
        """Set zoom factor."""
        if not 0.25 <= value <= 2.0:
            raise ValueError("Zoom factor must between 0.25 and 2.0.")
        # To keep table at the same position.
        zoom_ratio = 1 / self._zoom * value
        pos = self.verticalScrollBar().sliderPosition()
        self.verticalScrollBar().setSliderPosition(int(pos * zoom_ratio))
        pos = self.horizontalScrollBar().sliderPosition()
        self.horizontalScrollBar().setSliderPosition(int(pos * zoom_ratio))

        # Zoom font size
        font = self.font()
        font.setPointSize(int(self._initial_font_size * value))
        self.setFont(font)
        self.verticalHeader().setFont(font)
        self.horizontalHeader().setFont(font)

        # Zoom section size of headers
        h, v = self._initial_section_size
        self.setSectionSize(int(h * value), int(v * value))

        # Update stuff
        self._zoom = value
        return

    def wheelEvent(self, e: QtGui.QWheelEvent) -> None:
        """Zoom in/out table."""
        if e.modifiers() & Qt.ControlModifier:
            dt = e.angleDelta().y() / 120
            zoom = self.zoom() + 0.15 * dt
            self.setZoom(min(max(zoom, 0.25), 2.0))
            return None

        return super().wheelEvent(e)

    def sectionSize(self) -> tuple[int, int]:
        return (
            self.horizontalHeader().defaultSectionSize(),
            self.verticalHeader().defaultSectionSize(),
        )

    def setSectionSize(self, horizontal: int, vertical: int) -> None:
        """Update section size of headers."""
        self.verticalHeader().setDefaultSectionSize(vertical)
        self.horizontalHeader().setDefaultSectionSize(horizontal)
        return


class QBaseTable(QtW.QWidget):
    """
    The base widget for a table.

    Abstract Methods
    ----------------
    def createQTableView(self) -> None: ...
    def getDataFrame(self) -> pd.DataFrame: ...
    def setDataFrame(self) -> None: ...
    def createModel(self) -> AbstractDataFrameModel: ...
    def tableSlice(self) -> pd.DataFrame: ...
    """

    selectionChangedSignal = Signal(list)
    _DEFAULT_EDITABLE = False

    def __init__(
        self, parent: QtW.QWidget | None = None, data: pd.DataFrame | None = None
    ):
        super().__init__(parent)
        self._filter_slice: FilterType | None = None
        self.createQTableView()
        self.createModel()
        self.setDataFrame(data)
        self._qtable_view.setVerticalScrollMode(_SCROLL_PER_PIXEL)
        self._qtable_view.setHorizontalScrollMode(_SCROLL_PER_PIXEL)
        self._qtable_view.setFrameStyle(QtW.QFrame.Shape.NoFrame)
        delegate = TableItemDelegate(parent=self)
        self._qtable_view.setItemDelegate(delegate)
        self._qtable_view.selectionChangedSignal.connect(
            lambda: self.selectionChangedSignal.emit(self.selections())
        )
        if not self._DEFAULT_EDITABLE:
            self._qtable_view.setEditTriggers(_READ_ONLY)

    @property
    def _qtable_view(self) -> _QTableViewEnhanced:
        raise NotImplementedError()

    def createQTableView(self) -> None:
        """Create QTableView."""
        raise NotImplementedError()

    def getDataFrame(self) -> pd.DataFrame:
        raise NotImplementedError()

    def setDataFrame(self) -> None:
        raise NotImplementedError()

    def createModel(self) -> AbstractDataFrameModel:
        raise NotImplementedError()

    def tableSlice(self) -> pd.DataFrame:
        raise NotImplementedError()

    def tableShape(self) -> tuple[int, int]:
        model = self._qtable_view.model()
        nr = model.rowCount()
        nc = model.columnCount()
        return (nr, nc)

    def zoom(self) -> float:
        """Get current zoom factor."""
        return self._qtable_view.zoom()

    def setZoom(self, value: float) -> None:
        """Set zoom factor."""
        return self._qtable_view.setZoom(value)

    def itemDelegate(self) -> TableItemDelegate:
        return QtW.QTableView.itemDelegate(self._qtable_view)

    def model(self) -> AbstractDataFrameModel:
        return QtW.QTableView.model(self._qtable_view)

    def precision(self) -> int:
        """Return table value precision."""
        return self.itemDelegate().ndigits

    def setPrecision(self, ndigits: int) -> None:
        """Set table value precision."""
        ndigits = int(ndigits)
        if ndigits <= 0:
            raise ValueError("Cannot set negative precision.")
        self.itemDelegate().ndigits = ndigits

    def connectSelectionChangedSignal(self, slot):
        self.selectionChangedSignal.connect(slot)
        return slot

    def selections(self) -> Selections:
        """Get list of selections as slicable tuples"""
        qtable = self._qtable_view
        selections = qtable.selectionModel().selection()

        # selections = self.selectedRanges()
        out: Selections = []
        for i in range(len(selections)):
            sel = selections[i]
            r0 = sel.top()
            r1 = sel.bottom() + 1
            c0 = sel.left()
            c1 = sel.right() + 1
            out.append((slice(r0, r1), slice(c0, c1)))

        return out

    def setSelections(self, selections: Selections):
        """Set list of selections."""
        qtable = self._qtable_view
        qtable.clearSelection()

        model = self.model()
        nr, nc = model.df.shape
        try:
            for sel in selections:
                r, c = sel
                r0, r1, _ = r.indices(nr)
                c0, c1, _ = c.indices(nc)
                selection = QtCore.QItemSelection(
                    model.index(r0, c0), model.index(r1 - 1, c1 - 1)
                )
                qtable.selectionModel().select(
                    selection, QtCore.QItemSelectionModel.SelectionFlag.Select
                )

        except Exception as e:
            qtable.clearSelection()
            raise e

    def copyToClipboard(self, headers: bool = True):
        """Copy currently selected cells to clipboard."""
        selections = self.selections()
        if len(selections) == 0:
            return
        r_ranges = set()
        c_ranges = set()
        for rsel, csel in selections:
            r_ranges.add((rsel.start, rsel.stop))
            c_ranges.add((csel.start, csel.stop))

        nr = len(r_ranges)
        nc = len(c_ranges)
        if nr > 1 and nc > 1:
            show_messagebox(
                mode="error", title="Error", text="Wrong selection range.", parent=self
            )
            return
        else:
            data = self.model().df
            if nr == 1:
                axis = 1
            else:
                axis = 0
            ref = pd.concat([data.iloc[sel] for sel in selections], axis=axis)
            ref.to_clipboard(index=headers, header=headers)

    def readClipBoard(self) -> pd.DataFrame:
        """Read clipboard data and return as pandas DataFrame."""
        return pd.read_clipboard(header=None)

    def keyPressEvent(self, e: QtGui.QKeyEvent):
        _mod = e.modifiers()
        _key = e.key()
        if _mod & Qt.ControlModifier and _key == Qt.Key.Key_C:
            headers = _mod & Qt.ShiftModifier
            return self.copyToClipboard(headers)

        return super().keyPressEvent(e)

    def filter(self) -> FilterType | None:
        """Return the current filter."""
        return self._filter_slice

    def setFilter(self, sl: FilterType):
        """Set filter to the table view."""
        # NOTE: This method is also called when table needs initialization.

        self._filter_slice = sl
        data_sliced = self.tableSlice()

        if sl is None:
            self.model().df = data_sliced
        else:
            try:
                if callable(sl):
                    sl_filt = sl(data_sliced)
                else:
                    sl_filt = sl
                self.model().df = data_sliced[sl_filt]
            except Exception as e:
                self._filter_slice = None
                msg = f"Error in filter.\n\n{type(e).__name__} {e}\n\n Filter is reset."
                show_messagebox("error", "Error", msg, self)
        self.refresh()

    def refresh(self) -> None:
        qtable = self._qtable_view
        qtable.viewport().update()
        # headers have also to be updated.
        qtable.horizontalHeader().viewport().update()
        qtable.verticalHeader().viewport().update()
        return None


class QMutableTable(QBaseTable):
    itemChangedSignal = Signal(ItemInfo)
    selectionChangedSignal = Signal(list)
    _data_raw: pd.DataFrame

    def __init__(
        self, parent: QtW.QWidget | None = None, data: pd.DataFrame | None = None
    ):
        super().__init__(parent, data)
        self._editable = False
        self.model().dataEdited.connect(self.setDataFrameValue)

        # header editing signals
        self._qtable_view.horizontalHeader().sectionDoubleClicked.connect(
            self.editHorizontalHeader
        )
        self._qtable_view.verticalHeader().sectionDoubleClicked.connect(
            self.editVerticalHeader
        )

    def tableShape(self) -> tuple[int, int]:
        model = self.model()
        nr = model.rowCount()
        nc = model.columnCount()
        return (nr, nc)

    def tableSlice(self) -> pd.DataFrame:
        # TODO: just for now!!
        return self._data_raw

    def convertValue(self, r: int, c: int, value: Any) -> Any:
        """Convert value before updating DataFrame."""
        return value

    @overload
    def setDataFrameValue(self, r: int, c: int, value: Any) -> None:
        ...

    @overload
    def setDataFrameValue(self, r: slice, c: slice, value: pd.DataFrame) -> None:
        ...

    def setDataFrameValue(self, r, c, value) -> None:
        data = self._data_raw

        # convert values
        if isinstance(r, int) and isinstance(c, int):
            _value = self.convertValue(r, c, value)
        elif isinstance(r, slice) and isinstance(c, slice):
            _value: pd.DataFrame = value
            if _value.size == 1:
                v = _value.values[0, 0]
                _value = data.iloc[r, c].copy()
                for _ir, _r in enumerate(range(r.start, r.stop)):
                    for _ic, _c in enumerate(range(c.start, c.stop)):
                        _value.iloc[_ir, _ic] = self.convertValue(_r, _c, v)
            else:
                for _ir, _r in enumerate(range(r.start, r.stop)):
                    for _ic, _c in enumerate(range(c.start, c.stop)):
                        _value.iloc[_ir, _ic] = self.convertValue(
                            _r, _c, _value.iloc[_ir, _ic]
                        )
        else:
            raise TypeError

        if self._filter_slice is None:
            r0 = r
        else:
            if callable(self._filter_slice):
                sl = self._filter_slice(data)
            else:
                sl = self._filter_slice

            spec = np.where(sl)[0].tolist()
            r0 = spec[r]
            self.model().updateValue(r, c, _value)
        _old_value = data.iloc[r0, c]
        data.iloc[r0, c] = _value

        # emit item changed signal if value changed
        # NOTE pd.NA == x returns pd.NA, not False
        emit = False
        if isinstance(_value, pd.DataFrame):
            emit = True

        elif pd.isna(_value):
            if not pd.isna(_old_value):
                emit = True
        else:
            if pd.isna(_old_value) or _value != _old_value:
                emit = True

        if emit:
            self.itemChangedSignal.emit(ItemInfo(r, c, _value, _old_value))
        self.refresh()
        return None

    def editability(self) -> bool:
        """Return the editability of the table."""
        return self._editable

    def setEditability(self, editable: bool):
        """Set the editability of the table."""
        if editable:
            self._qtable_view.setEditTriggers(_EDITABLE)
        else:
            self._qtable_view.setEditTriggers(_READ_ONLY)
        self._editable = editable

    def connectItemChangedSignal(self, slot):
        self.itemChangedSignal.connect(slot)
        return slot

    def keyPressEvent(self, e: QtGui.QKeyEvent):
        _mod = e.modifiers()
        _key = e.key()
        if _mod & Qt.ControlModifier and _key == Qt.Key.Key_C:
            headers = _mod & Qt.ShiftModifier
            return self.copyToClipboard(headers)
        elif _mod & Qt.ControlModifier and _key == Qt.Key.Key_V:
            return self.pasteFromClipBoard()
        elif _mod == Qt.NoModifier and _key in (
            Qt.Key.Key_Delete,
            Qt.Key.Key_Backspace,
        ):
            return self.deleteValues()
        elif _mod in (Qt.NoModifier, Qt.ShiftModifier) and (
            Qt.Key.Key_Exclam <= _key <= Qt.Key.Key_ydiaeresis
        ):
            # Enter editing mode
            qtable = self._qtable_view
            text = QtGui.QKeySequence(_key).toString()
            if _mod != Qt.ShiftModifier:
                text = text.lower()
            self.model().setData(qtable.currentIndex(), text, Qt.ItemDataRole.EditRole)
            qtable.edit(qtable.currentIndex())
            focused_widget = QtW.QApplication.focusWidget()
            if isinstance(focused_widget, QtW.QLineEdit):
                focused_widget.deselect()
            return

        return super().keyPressEvent(e)

    def pasteFromClipBoard(self):
        """
        Paste data to table.

        This function supports many types of pasting.
        1. Single selection, single data in clipboard -> just paste
        2. Single selection, multiple data in clipboard -> paste starts from the selection position.
        3. Multiple selection, single data in clipboard -> paste the same value for all the selection.
        4. Multiple selection, multiple data in clipboard -> paste only if their shape is identical.

        Also, if data is filtrated, pasted data also follows the filtration.
        """
        selections = self.selections()
        n_selections = len(selections)
        if n_selections == 0 or not self.editability():
            return
        elif n_selections > 1:
            return show_messagebox(
                mode="error",
                title="Error",
                text="Cannot paste with multiple selections.",
                parent=self,
            )

        df = self.readClipBoard()

        # check size and normalize selection slices
        sel = selections[0]
        rrange, crange = sel
        rlen = rrange.stop - rrange.start
        clen = crange.stop - crange.start
        dr, dc = df.shape
        size = dr * dc

        if rlen * clen == 1 and size > 1:
            sel = (
                slice(rrange.start, rrange.start + dr),
                slice(crange.start, crange.start + dc),
            )

        elif size > 1 and (rlen, clen) != (dr, dc):
            # If selection is column-wide or row-wide, resize them
            model = self.model()
            if rlen == model.df.shape[0]:
                rrange = slice(0, dr)
                rlen = dr
            if clen == model.df.shape[1]:
                crange = slice(0, dc)
                clen = dc

            if (rlen, clen) != (dr, dc):
                return show_messagebox(
                    mode="error",
                    title="Error",
                    text=f"Shape mismatch between data in clipboard {(rlen, clen)} and "
                    f"destination {(dr, dc)}.",
                    parent=self,
                )
            else:
                sel = (rrange, crange)

        rsel, csel = sel

        # check dtype
        dtype_src = df.dtypes
        dtype_dst = self._data_raw.dtypes[csel]
        if any(a.kind != b.kind for a, b in zip(dtype_src.values, dtype_dst.values)):
            return show_messagebox(
                mode="error",
                title="Error",
                text=f"Data type mismatch between data in clipboard {list(dtype_src)} and "
                f"destination {list(dtype_dst)}.",
                parent=self,
            )

        # update table
        try:
            self.setDataFrameValue(rsel, csel, df)

        except Exception as e:
            show_messagebox(
                mode="error",
                title=e.__class__.__name__,
                text=str(e),
                parent=self,
            )
            raise e from None

        else:
            self.setSelections([sel])

        return None

    def deleteValues(self):
        """Replace selected cells with NaN."""
        selections = self.selections()
        for sel in selections:
            rsel, csel = sel
            nr = rsel.stop - rsel.start
            nc = csel.stop - csel.start
            dtypes = list(self._data_raw.dtypes[csel])
            df = pd.DataFrame(
                {c: pd.Series(np.full(nr, np.nan), dtype=dtypes[c]) for c in range(nc)},
            )
            self.setDataFrameValue(rsel, csel, df)

    def editHorizontalHeader(self, index: int):
        """Edit the horizontal header."""
        if not self.editability():
            return

        qtable = self._qtable_view
        _header = qtable.horizontalHeader()
        _line = QtW.QLineEdit(_header)
        edit_geometry = _line.geometry()
        edit_geometry.setHeight(_header.height())
        edit_geometry.setWidth(_header.sectionSize(index))
        edit_geometry.moveLeft(_header.sectionViewportPosition(index))
        _line.setGeometry(edit_geometry)
        _line.setHidden(False)
        _line.setAlignment(Qt.AlignmentFlag.AlignCenter)

        column_axis = self.model().df.columns
        if index < column_axis.size:
            text = column_axis[index]
        else:
            text = ""

        _line.setText(str(text))
        _line.setFocus()

        self._line = _line

        @_line.editingFinished.connect
        def _set_header_data():
            self._line.setHidden(True)
            self.setHorizontalHeaderValue(index, self._line.text())
            return None

        return None

    def editVerticalHeader(self, index: int):
        if not self.editability():
            return

        qtable = self._qtable_view
        _header = qtable.verticalHeader()
        _line = QtW.QLineEdit(_header)
        edit_geometry = _line.geometry()
        edit_geometry.setHeight(_header.sectionSize(index))
        edit_geometry.setWidth(_header.width())
        edit_geometry.moveTop(_header.sectionViewportPosition(index))
        _line.setGeometry(edit_geometry)
        _line.setHidden(False)

        index_axis = self.model().df.index

        if index < index_axis.size:
            text = index_axis[index]
        else:
            text = ""

        _line.setText(str(text))
        _line.setFocus()
        _line.setAlignment(Qt.AlignmentFlag.AlignCenter)

        self._line = _line

        @_line.editingFinished.connect
        def _set_header_data():
            self._line.setHidden(True)
            self.setVerticalHeaderValue(index, self._line.text())

    def setHorizontalHeaderValue(self, index: int, value: Any) -> None:
        qtable = self._qtable_view
        column_axis = self.model().df.columns
        _header = qtable.horizontalHeader()

        mapping = {column_axis[index]: value}

        self._data_raw.rename(columns=mapping, inplace=True)
        self.model().df.rename(columns=mapping, inplace=True)

        size_hint = _header.sectionSizeHint(index)
        if _header.sectionSize(index) < size_hint:
            _header.resizeSection(index, size_hint)
        return None

    def setVerticalHeaderValue(self, index: int, value: Any) -> None:
        qtable = self._qtable_view
        index_axis = self.model().df.index
        _header = qtable.verticalHeader()

        mapping = {index_axis[index]: value}

        self._data_raw.rename(index=mapping, inplace=True)
        self.model().df.rename(index=mapping, inplace=True)
        _width_hint = _header.sizeHint().width()
        _header.resize(QtCore.QSize(_width_hint, _header.height()))
        return None


class QMutableSimpleTable(QMutableTable):
    @property
    def _qtable_view(self) -> _QTableViewEnhanced:
        return self._qtable_view_

    def createQTableView(self):
        self._qtable_view_ = _QTableViewEnhanced()
        _layout = QtW.QVBoxLayout()
        _layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(_layout)
        self.layout().addWidget(self._qtable_view_)


class TableItemDelegate(QtW.QStyledItemDelegate):
    """Displays table widget items with properly formatted numbers."""

    def __init__(self, parent: QtCore.QObject | None = None, ndigits: int = 4) -> None:
        super().__init__(parent)
        self.ndigits = ndigits

    def displayText(self, value, locale):
        return super().displayText(self._format_number(value), locale)

    def createEditor(
        self, parent: QtW.QWidget, option, index: QtCore.QModelIndex
    ) -> QtW.QWidget:
        """Create different type of editors for different dtypes."""
        table = parent.parent().parent()
        if isinstance(table, QMutableTable):
            df = table.model().df
            row = index.row()
            col = index.column()
            if row >= df.shape[0] or col >= df.shape[1]:
                # out of bounds
                return super().createEditor(parent, option, index)

            dtype: np.dtype = df.dtypes[col]
            if dtype == "category":
                # use combobox for categorical data
                cbox = QtW.QComboBox(parent)
                choices = list(map(str, dtype.categories))
                cbox.addItems(choices)
                cbox.setCurrentIndex(choices.index(df.iat[row, col]))
                return cbox
            elif dtype == "bool":
                # use checkbox for boolean data
                cbox = QtW.QComboBox(parent)
                choices = ["True", "False"]
                cbox.addItems(choices)
                cbox.setCurrentIndex(0 if df.iat[row, col] else 1)
                return cbox

        return super().createEditor(parent, option, index)

    def setEditorData(self, editor: QtW.QWidget, index: QtCore.QModelIndex) -> None:
        super().setEditorData(editor, index)
        if isinstance(editor, QtW.QComboBox):
            editor.showPopup()
        return None

    # modified from magicgui
    def _format_number(self, text: str) -> str:
        """convert string to int or float if possible"""
        try:
            value: int | float | None = int(text)
        except ValueError:
            try:
                value = float(text)
            except ValueError:
                value = None

        ndigits = self.ndigits

        if isinstance(value, (int, float)):
            if 0.1 <= abs(value) < 10 ** (ndigits + 1) or value == 0:
                text = str(value) if isinstance(value, int) else f"{value:.{ndigits}f}"
            else:
                text = f"{value:.{ndigits-1}e}"

        return text



from __future__ import annotations
from typing import TYPE_CHECKING, Callable
from qtpy import QtWidgets as QtW
from qtpy.QtCore import Qt, QTimer
from ._base import QBaseTable, _QTableViewEnhanced, DataFrameModel

if TYPE_CHECKING:
    import pandas as pd


def _get_standard_icon(x):
    return QtW.QApplication.style().standardIcon(x)


class QPlayButton(QtW.QToolButton):
    def __init__(self, parent=None):
        super().__init__(parent)
        self._RUN = _get_standard_icon(QtW.QStyle.StandardPixmap.SP_MediaPlay)
        self._PAUSE = _get_standard_icon(QtW.QStyle.StandardPixmap.SP_MediaPause)
        self.clicked.connect(self.switchRunning)
        self.setRunning(True)

    def running(self):
        return self._running

    def setRunning(self, val: bool):
        self._running = val
        if val:
            self.setIcon(self._PAUSE)
        else:
            self.setIcon(self._RUN)

    def switchRunning(self, *_):
        self.setRunning(not self.running())


class QTableDisplay(QBaseTable):
    def __init__(
        self,
        parent: QtW.QWidget | None = None,
        loader: Callable[[], pd.DataFrame] = None,
        interval_ms: int = 1000,
    ):
        import pandas as pd

        super().__init__(parent, pd.DataFrame([]))
        if loader is None:
            self._loader = lambda: pd.DataFrame([])
        else:
            self._loader = lambda: pd.DataFrame(loader())
        self._refreshing = False
        self._qtimer = QTimer()
        self._qtimer.setSingleShot(True)
        self._qtimer.setInterval(interval_ms)
        self._qtimer.setTimerType(Qt.TimerType.PreciseTimer)
        self._qtimer.timeout.connect(self.refreshCallback)

        self._interval_spinbox.valueChanged.connect(self._qtimer.setInterval)
        self._interval_spinbox.setValue(self._qtimer.interval())

        if self._play_button.running():
            self._qtimer.start()

    if TYPE_CHECKING:

        def model(self) -> DataFrameModel:
            ...

    def refreshCallback(self):
        """Run refresh if needed."""
        if self._play_button.running() and not self._refreshing:
            self.refresh()

    def loader(self) -> Callable:
        """Return the loader function."""
        return self._loader

    def setLoader(self, loader: Callable) -> None:
        """Set the loader function and refresh."""
        if not callable(loader):
            raise TypeError("loader must be callable")
        self._loader = loader
        self.refresh()

    def getDataFrame(self) -> pd.DataFrame:
        return self._data_raw

    def setDataFrame(self, data: pd.DataFrame) -> None:
        self._data_raw = data
        self.model().df = data
        self._qtable_view.viewport().update()
        return

    def createModel(self):
        model = DataFrameModel(self)
        self._qtable_view.setModel(model)
        return None

    @property
    def _qtable_view(self) -> _QTableViewEnhanced:
        return self._qtable_view_

    def createQTableView(self):
        self._qtable_view_ = _QTableViewEnhanced()

        _header = QtW.QWidget()
        _header_layout = QtW.QHBoxLayout()
        _header_layout.setContentsMargins(2, 2, 2, 2)
        _header_layout.addWidget(QtW.QLabel("Interval (ms):"))

        self._interval_spinbox = QtW.QSpinBox()
        self._interval_spinbox.setMinimum(10)
        self._interval_spinbox.setMaximum(10000)
        _header_layout.addWidget(self._interval_spinbox)

        self._play_button = QPlayButton()

        @self._play_button.clicked.connect
        def _clicked(*_):
            if self._qtimer.isActive():
                self._qtimer.stop()
            else:
                self._qtimer.start()

        _header_layout.addWidget(self._play_button)

        _header.setLayout(_header_layout)

        _layout = QtW.QVBoxLayout()
        _layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(_layout)
        self.layout().addWidget(_header)
        self.layout().addWidget(self._qtable_view_)

    def refresh(self) -> None:
        self._refreshing = True
        if self.isVisible():
            try:
                self._data_raw = self._loader()
            except Exception as e:
                return
            self.model().df = self._data_raw

        self._refreshing = False
        if self._play_button.running():
            self._qtimer.start()
        else:
            self._qtimer.stop()
        return super().refresh()



from __future__ import annotations
from typing import Any
from io import StringIO
import numpy as np
import pandas as pd
from qtpy import QtCore, QtWidgets as QtW
from qtpy.QtCore import Qt

from ._base import AbstractDataFrameModel, QMutableSimpleTable

MAX_ROW_SIZE = 100000
MAX_COLUMN_SIZE = 100000


class SpreadSheetModel(AbstractDataFrameModel):
    @property
    def df(self) -> pd.DataFrame:
        return self._df

    @df.setter
    def df(self, data: pd.DataFrame):
        self._df = data

    def rowCount(self, parent=None):
        return min(self._df.shape[0] + 10, MAX_ROW_SIZE)

    def columnCount(self, parent=None):
        return min(self._df.shape[1] + 10, MAX_COLUMN_SIZE)

    def insertRows(
        self, row: int, count: int, parent: QtCore.QModelIndex = None
    ) -> bool:
        df = self.df
        self.beginInsertRows(parent, row, row + count - 1)
        df0 = df.iloc[:row, :]
        df1 = pd.DataFrame(
            np.full((count, df.shape[1]), np.nan), columns=df.columns, dtype="string"
        )
        df2 = df.iloc[row:, :]
        self.df = pd.concat([df0, df1, df2], axis=0)
        self.endInsertRows()
        return True

    def insertColumns(
        self, column: int, count: int, parent: QtCore.QModelIndex = None
    ) -> bool:
        df = self.df
        self.beginInsertColumns(parent, column, column + count - 1)
        df0 = df.iloc[:, :column]
        df1 = pd.DataFrame(
            np.full((df.shape[0], count), np.nan), index=df.index, dtype="string"
        )
        df2 = df.iloc[:, column:]
        self.df = pd.concat([df0, df1, df2], axis=1)
        self.endInsertColumns()
        return True

    def removeRows(
        self, row: int, count: int, parent: QtCore.QModelIndex = None
    ) -> bool:
        df = self.df
        self.beginRemoveRows(parent, row, row + count - 1)
        self.df = df.drop(index=df.index[row])
        self.endRemoveRows()
        return True

    def removeColumns(
        self, column: int, count: int, parent: QtCore.QModelIndex = None
    ) -> bool:
        df = self.df
        self.beginRemoveColumns(parent, column, column + count - 1)
        self.df = df.drop(columns=df.columns[column])
        self.endRemoveColumns()
        return True


class QSpreadSheet(QMutableSimpleTable):
    """
    A table layer class that works similar to Excel sheet.

    Unlike ``QTableLayer``, this class does not have dtype. The dtype will be
    determined every time table data is converted into DataFrame. Table data
    is (almost) unbounded.
    """

    _DEFAULT_EDITABLE = True

    def __init__(self, parent=None, data: pd.DataFrame | None = None):
        super().__init__(parent, data)
        self._data_cache = None
        self._qtable_view.rightClickedSignal.connect(self.showContextMenu)

    def getDataFrame(self) -> pd.DataFrame:
        if self._data_cache is not None:
            return self._data_cache
        # Convert table data into a DataFrame with the optimal dtypes
        buf = StringIO(self._data_raw.to_csv(sep="\t", index_label="_INDEX_"))
        out = pd.read_csv(buf, sep="\t", index_col="_INDEX_")
        out.index.name = None
        self._data_cache = out
        return out

    def setDataFrame(self, data: pd.DataFrame) -> None:
        self._data_raw = data.astype("string")
        self.model().setShape(data.index.size + 10, data.columns.size + 10)
        self._data_cache = None
        self.setFilter(None)
        self.refresh()
        return

    def createModel(self) -> None:
        """Create spreadsheet model."""
        model = SpreadSheetModel(self)
        self._qtable_view.setModel(model)
        return None

    def convertValue(self, r: int, c: int, value: Any) -> Any:
        """Convert value to the type of the table."""
        return value

    def readClipBoard(self):
        return pd.read_clipboard(header=None).astype("string")

    def setDataFrameValue(self, r: int | slice, c: int | slice, value: Any) -> None:
        nr, nc = self._data_raw.shape
        rmax = _get_limit(r)
        cmax = _get_limit(c)
        if nr <= rmax or nc <= cmax:
            if nr <= rmax:
                self.expandRows(rmax - nr + 1)
            if self._data_raw.shape[1] <= cmax:  # NOTE: DataFrame shape is updated
                self.expandColumns(cmax - nc + 1)
            new_shape = self._data_raw.shape
            self.model().setShape(new_shape[0] + 10, new_shape[1] + 10)

        self._data_cache = None
        super().setDataFrameValue(r, c, value)

        self.setFilter(self._filter_slice)
        self._qtable_view.verticalHeader().resize(
            self._qtable_view.verticalHeader().sizeHint()
        )
        return None

    def expandRows(self, n_expand: int):
        if self._data_raw.size == 0:
            self._data_raw = pd.DataFrame(
                np.full((n_expand, 1), np.nan),
                index=range(n_expand),
                dtype="string",
            )
            return
        nr, nc = self._data_raw.shape
        ext = pd.DataFrame(
            np.full((n_expand, nc), np.nan),
            index=range(nr, n_expand + nr),
            columns=self._data_raw.columns,
            dtype="string",
        )
        self._data_raw = pd.concat([self._data_raw, ext], axis=0)
        return None

    def expandColumns(self, n_expand: int):
        if self._data_raw.size == 0:
            self._data_raw = pd.DataFrame(
                np.full((1, n_expand), np.nan),
                columns=range(n_expand),
                dtype="string",
            )
            return
        nr, nc = self._data_raw.shape
        ext = pd.DataFrame(
            np.full((nr, n_expand), np.nan),
            index=self._data_raw.index,
            columns=range(nc, n_expand + nc),
            dtype="string",
        )
        self._data_raw = pd.concat([self._data_raw, ext], axis=1)
        return None

    def setVerticalHeaderValue(self, index: int, value: Any) -> None:
        """Set value of the table vertical header and DataFrame at the index."""
        nrows = self._data_raw.shape[0]
        if index >= nrows:
            if value == "":
                return
            self.expandRows(index - nrows + 1)
            self.setFilter(self._filter_slice)

        new_shape = self._data_raw.shape
        self._data_cache = None
        self.setFilter(self._filter_slice)
        self.model().setShape(new_shape[0] + 10, new_shape[1] + 10)

        return super().setVerticalHeaderValue(index, value)

    def setHorizontalHeaderValue(self, index: int, value: Any) -> None:
        """Set value of the table horizontal header and DataFrame at the index."""
        ncols = self._data_raw.shape[1]
        if index >= ncols:
            if value == "":
                return
            self.expandColumns(index - ncols + 1)
            self.setFilter(self._filter_slice)

        new_shape = self._data_raw.shape
        self._data_cache = None
        self.setFilter(self._filter_slice)
        self.model().setShape(new_shape[0] + 10, new_shape[1] + 10)

        return super().setHorizontalHeaderValue(index, value)

    def showContextMenu(self, pos: QtCore.QPoint):
        menu = QtW.QMenu(self._qtable_view)
        index = self._qtable_view.indexAt(pos)
        row, col = index.row(), index.column()
        model = self.model()
        menu.addAction(
            "Insert a row above", lambda: model.insertRows(row, 1, QtCore.QModelIndex())
        )
        menu.addAction(
            "Insert a row below",
            lambda: model.insertRows(row + 1, 1, QtCore.QModelIndex()),
        )
        menu.addAction(
            "Insert a column on the left",
            lambda: model.insertColumns(col, 1, QtCore.QModelIndex()),
        )
        menu.addAction(
            "Insert a column on the right",
            lambda: model.insertColumns(col + 1, 1, QtCore.QModelIndex()),
        )
        menu.addAction(
            "Remove this row", lambda: model.removeRows(row, 1, QtCore.QModelIndex())
        )
        menu.addAction(
            "Remove this column",
            lambda: model.removeColumns(col, 1, QtCore.QModelIndex()),
        )
        return menu.exec(self._qtable_view.mapToGlobal(pos))


def _get_limit(a) -> int:
    if isinstance(a, int):
        amax = a
    elif isinstance(a, slice):
        amax = a.stop - 1
    else:
        raise TypeError(f"Cannot infer limit of type {type(a)}")
    return amax



from __future__ import annotations
from typing import Any, Callable, Iterable, TYPE_CHECKING, TypeVar
from qtpy.QtWidgets import QWidget, QVBoxLayout
from magicgui import register_type
from magicgui.widgets import Widget, Container, ComboBox, Label, Dialog
from magicgui.widgets._bases import CategoricalWidget
from magicgui.backends._qtpy.widgets import QBaseWidget

from .widgets import TableViewer, TableView, TableViewerWidget
from .types import (
    TableColumn,
    TableData,
    TableDataTuple,
    TableInfoInstance,
    TabPosition,
)

if TYPE_CHECKING:
    import pandas as pd

# #############################################################################
#    magicgui-widget
# #############################################################################


class MagicTableViewer(Widget, TableViewerWidget):
    """
    A magicgui widget of table viewer.

    This class is a subclass of ``magicgui.widget.Widget`` so that it can be used
    in a compatible way with magicgui and napari.

    Parameters
    ----------
    tab_position: TabPosition or str
        Type of list-like widget to use.
    """

    def __init__(
        self,
        *,
        tab_position: TabPosition | str = TabPosition.top,
        name: str = "",
        label: str = None,
        tooltip: str | None = None,
        visible: bool | None = None,
        enabled: bool = True,
    ):
        super().__init__(
            widget_type=QBaseWidget,
            backend_kwargs={"qwidg": QWidget},
            name=name,
            label=label,
            tooltip=tooltip,
            visible=visible,
            enabled=enabled,
        )
        TableViewerWidget.__init__(self, tab_position=tab_position, show=False)
        self.native: QWidget
        self.native.setLayout(QVBoxLayout())
        self.native.layout().addWidget(self._qwidget)
        self.native.setContentsMargins(0, 0, 0, 0)


# #############################################################################
#    magicgui type registration
# #############################################################################

_DEFAULT_NAME = "Result"


def find_table_viewer_ancestor(widget: Widget | QWidget) -> TableViewer | None:
    from ._qt._mainwindow import _QtMainWidgetBase

    if isinstance(widget, Widget):
        qwidget = widget.native
    elif isinstance(widget, QWidget):
        qwidget = widget
    else:
        raise TypeError(f"Cannot use {type(widget)} as an input.")
    qwidget: QWidget
    parent = qwidget.parent()
    while (parent := qwidget.parent()) is not None:
        qwidget = parent
        if isinstance(qwidget, _QtMainWidgetBase):
            return qwidget._table_viewer

    return None


def get_tables(widget: CategoricalWidget) -> list[tuple[str, Any]]:
    v = find_table_viewer_ancestor(widget)
    if v is None:
        return []
    return v.tables


def get_table_data(widget: CategoricalWidget) -> list[tuple[str, Any]]:
    v = find_table_viewer_ancestor(widget)
    if v is None:
        return []
    return [(table.name, table.data) for table in v.tables]


def open_viewer(gui, result: TableViewer, return_type: type):
    result.show()


def add_table_to_viewer(gui, result: Any, return_type: type):
    viewer = find_table_viewer_ancestor(gui)
    if viewer is None:
        return
    viewer.add_layer(result)


def add_table_data_to_viewer(gui, result: Any, return_type: type):
    viewer = find_table_viewer_ancestor(gui)
    if viewer is None:
        return
    viewer.add_table(result, name=_DEFAULT_NAME)


def add_table_data_tuple_to_viewer(gui, result: tuple, return_type: type):
    viewer = find_table_viewer_ancestor(gui)
    if viewer is None:
        return
    n = len(result)
    if n == 1:
        data = (result[0], _DEFAULT_NAME, {})
    elif n == 2:
        if isinstance(result[1], dict):
            name = result[1].pop("name", _DEFAULT_NAME)
            data = (result[0], name, result[1])
        else:
            data = (result[0], result[1], {})
    elif n == 3:
        data = result
    else:
        raise ValueError(f"Length of TableDataTuple must be < 4, got {n}.")
    viewer.add_table(data[0], name=data[1], **data[2])


register_type(
    TableViewer, return_callback=open_viewer, choices=find_table_viewer_ancestor
)
register_type(TableView, return_callback=add_table_to_viewer, choices=get_tables)
register_type(
    TableData,
    return_callback=add_table_data_to_viewer,
    choices=get_table_data,
    nullable=False,
)
register_type(TableDataTuple, return_callback=add_table_data_tuple_to_viewer)

# Widget
class ColumnChoice(Container):
    def __init__(
        self,
        data_choices: Iterable[pd.DataFrame]
        | Callable[[Widget], Iterable[pd.DataFrame]],
        value=None,
        **kwargs,
    ):
        self._dataframe_choices = ComboBox(choices=data_choices, value=value, **kwargs)
        self._column_choices = ComboBox(choices=self._get_available_columns)
        _label_l = Label(value='["')
        _label_l.max_width = 24
        _label_r = Label(value='"]')
        _label_r.max_width = 24
        super().__init__(
            layout="horizontal",
            widgets=[self._dataframe_choices, _label_l, self._column_choices, _label_r],
            labels=False,
            name=kwargs.get("name"),
        )
        self.margins = (0, 0, 0, 0)
        self._dataframe_choices.changed.connect(self._set_available_columns)

    def _get_available_columns(self, w=None):
        df: pd.DataFrame = self._dataframe_choices.value
        cols = getattr(df, "columns", [])
        return cols

    def _set_available_columns(self, w=None):
        cols = self._get_available_columns()
        self._column_choices.choices = cols
        return None

    @property
    def value(self) -> pd.Series:
        df = self._dataframe_choices.value
        return df[self._column_choices.value]


register_type(
    TableColumn,
    widget_type=ColumnChoice,
    return_callback=add_table_data_to_viewer,
    data_choices=get_table_data,
    nullable=False,
)


class ColumnNameChoice(Container):
    """
    A container widget with a DataFrame selection and multiple column name selections.

    This widget is composed of two or more ComboBox widgets. The top one is to choose a
    DataFrame and the rest are to choose column names from the DataFrame. When the DataFrame
    selection changed, the column name selections will also changed accordingly.
    """

    def __init__(
        self,
        data_choices: Iterable[pd.DataFrame]
        | Callable[[Widget], Iterable[pd.DataFrame]],
        column_choice_names: Iterable[str],
        value=None,
        **kwargs,
    ):
        self._dataframe_choices = ComboBox(choices=data_choices, value=value, **kwargs)
        self._column_names: list[ComboBox] = []
        for cn in column_choice_names:
            self._column_names.append(
                ComboBox(choices=self._get_available_columns, name=cn, nullable=True)
            )
        self._child_container = Container(widgets=self._column_names, layout="vertical")
        self._child_container.margins = (0, 0, 0, 0)
        super().__init__(
            layout="vertical",
            widgets=[self._dataframe_choices, self._child_container],
            labels=False,
            name=kwargs.get("name"),
        )
        self.margins = (0, 0, 0, 0)
        self._dataframe_choices.changed.connect(self._set_available_columns)

    def _get_available_columns(self, w=None):
        df: pd.DataFrame = self._dataframe_choices.value
        cols = getattr(df, "columns", [])
        return cols

    def _set_available_columns(self, w=None):
        cols = self._get_available_columns()
        for cbox in self._column_names:
            cbox.choices = cols
        return None

    @property
    def value(self) -> tuple[pd.DataFrame, list[str]]:
        df = self._dataframe_choices.value
        colnames = [cbox.value for cbox in self._column_names]
        return (df, colnames)


register_type(
    TableInfoInstance, widget_type=ColumnNameChoice, data_choices=get_table_data
)

# #############################################################################
#    Utility functions
# #############################################################################

_F = TypeVar("_F", bound=Callable)


def dialog_factory(function: _F) -> _F:
    from magicgui.signature import magic_signature

    def _runner(parent=None, **param_options):
        widgets = list(
            magic_signature(function, gui_options=param_options).widgets().values()
        )
        dlg = Dialog(widgets=widgets)
        dlg.native.setParent(parent, dlg.native.windowFlags())
        if dlg.exec():
            out = function(**dlg.asdict())
        else:
            out = None
        return out

    return _runner



# from __future__ import annotations
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Union, List
import weakref
import numpy as np
import pandas as pd
from qtpy import QtWidgets as QtW, QtGui
from qtpy.QtWidgets import QAction

from ._svg import QColoredSVGIcon
from .._magicgui import dialog_factory
from ..types import TableData

if TYPE_CHECKING:
    from ._mainwindow import _QtMainWidgetBase
    from ..widgets.mainwindow import _TableViewerBase

SUMMARY_CHOICES = ["mean", "median", "std", "sem", "min", "max", "sum"]

ICON_DIR = Path(__file__).parent / "_icons"

class _QToolBar(QtW.QToolBar):
    def __init__(self, parent=None):
        super().__init__(parent)
        self._button_and_icon: List["tuple[QtW.QToolButton, QColoredSVGIcon]"] = []
    
    def updateIconColor(self, color):
        for button, icon in self._button_and_icon:
            button.setIcon(icon.colored(color))

class QTableStackToolBar(QtW.QToolBar):
    def __init__(self, parent: "_QtMainWidgetBase"):
        super().__init__(parent)
        
        self._tab = QtW.QTabWidget(self)
        self._tab.setContentsMargins(0, 0, 0, 0)
        self._tab.setStyleSheet(
            r"QTabWidget {margin: 0px, 0px, 0px, 0px; padding: 0px;}"
        )
        self._child_widgets: weakref.WeakValueDictionary[str, _QToolBar] = weakref.WeakValueDictionary()
        
        self.addWidget(self._tab)
        self.registerAction("File", self.open_table, ICON_DIR / "open_table.svg")
        self.registerAction("File", self.open_spreadsheet, ICON_DIR / "open_spreadsheet.svg")
        self.registerAction("File", self.save_table, ICON_DIR / "save_table.svg")
        
        self.registerAction("Table", self.copy_as_table, ICON_DIR / "copy_as_table.svg")
        self.registerAction("Table", self.copy_as_spreadsheet, ICON_DIR / "copy_as_spreadsheet.svg")
        self.registerAction("Table", self.new_spreadsheet, ICON_DIR / "new_spreadsheet.svg")
        self.addSeparatorToChild("Table")
        self.registerAction("Table", self.groupby, ICON_DIR / "groupby.svg")
        self.registerAction("Table", self.hconcat, ICON_DIR / "hconcat.svg")
        self.registerAction("Table", self.vconcat, ICON_DIR / "vconcat.svg")
        self.registerAction("Table", self.pivot, ICON_DIR / "pivot.svg")
        self.registerAction("Table", self.melt, ICON_DIR / "melt.svg")
        self.addSeparatorToChild("Table")
        self.registerAction("Table", self.query, ICON_DIR / "query.svg")
        
        self.registerAction("Analyze", self.summarize_table, ICON_DIR / "summarize_table.svg")
        self.addSeparatorToChild("Analyze")
        

    @property
    def viewer(self) -> "_TableViewerBase":
        return self.parent()._table_viewer

    if TYPE_CHECKING:
        def parent(self) -> "_QtMainWidgetBase":
            ...
    
    def addToolBar(self, name: str):
        if name in self._child_widgets.keys():
            raise ValueError(f"Tab with name {name!r} already exists.")
        toolbar = _QToolBar(self)
        toolbar.setContentsMargins(0, 0, 0, 0)
        self._tab.addTab(toolbar, name)
        self._child_widgets[name] = toolbar
        
    def registerAction(self, tabname: str, f: Callable, icon: Union[str, Path]):
        """Register a callback `f` in tab `tabname`."""
        if tabname not in self._child_widgets:
            self.addToolBar(tabname)
        toolbar = self._child_widgets[tabname]
        qicon = QColoredSVGIcon.fromfile(str(icon))
        action = toolbar.addAction(qicon, "")
        action.triggered.connect(f)
        action.setToolTip(f.__doc__)
        toolbar._button_and_icon.append((toolbar.widgetForAction(action), qicon))
        return None
    
    def addSeparatorToChild(self, tabname: str) -> QAction:
        toolbar = self._child_widgets[tabname]
        return toolbar.addSeparator()
    
    def setToolButtonColor(self, color: str):
        for toolbar in self._child_widgets.values():
            toolbar.updateIconColor(color)

    def open_table(self):
        """Open a file as a table."""
        return self._open(type="table")
        
    def open_spreadsheet(self):
        """Open a file as a spreadsheet."""
        return self._open(type="spreadsheet")
    
    def _open(self, type):
        # TODO: history
        paths, _ = QtW.QFileDialog.getOpenFileNames(self.parent(), "Open File", "")
        for path in paths:
            self.viewer.open(path, type=type)
        return None
        
    def save_table(self):
        """Save current table."""
        path, _ = QtW.QFileDialog.getSaveFileName(self.parent(), "Save Table", "")
        if path:
            self.viewer.save(path)
    
    def copy_as_table(self):
        """Make a copy of the current table."""
        viewer = self.viewer
        table = viewer.current_table
        viewer.add_table(table.data, name=f"{table.name}-copy")
    
    def copy_as_spreadsheet(self):
        """Make a copy of the current table."""
        viewer = self.viewer
        table = viewer.current_table
        viewer.add_spreadsheet(table.data, name=f"{table.name}-copy")
    
    def new_spreadsheet(self):
        """Create a new spreadsheet."""
        self.viewer.add_spreadsheet(name="New")

    def groupby(self):
        """Group table data by its column value."""
        table = self.viewer.current_table
        out = groupby(
            df={"bind": table.data},
            by={"choices": list(table.data.columns), "widget_type": "Select"},
        )
        if out is not None:
            self.viewer.add_groupby(out, name=f"{table.name}-groupby")
    
    def hconcat(self):
        out = hconcat(
            viewer={"bind": self.viewer},
            names={"value": [self.viewer.current_table.name],
                   "widget_type": "Select",
                   "choices": [t.name for t in self.viewer.tables]},
        )
        if out is not None:
            self.viewer.add_table(out, name=f"hconcat")
    
    def vconcat(self):
        out = vconcat(
            viewer={"bind": self.viewer},
            names={"value": [self.viewer.current_table.name],
                   "widget_type": "Select",
                   "choices": [t.name for t in self.viewer.tables]},
        )
        if out is not None:
            self.viewer.add_table(out, name=f"vconcat")
    
    def pivot(self):
        """Pivot a table."""
        table = self.viewer.current_table
        col = list(table.data.columns)
        if len(col) < 2:
            raise ValueError("Table must have at least two columns.")
        out = pivot(
            df={"bind": table.data},
            index={"choices": col, "value": col[0]},
            columns={"choices": col, "value": col[1]},
            values={"choices": col, "nullable": True},
        )
        if out is not None:
            self.viewer.add_table(out, name=f"{table.name}-pivot")
        
    def melt(self):
        """Unpivot a table."""
        table = self.viewer.current_table
        out = melt(
            df={"bind": table.data},
            id_vars={"choices": list(table.data.columns), "widget_type": "Select"},
        )
        if out is not None:
            self.viewer.add_table(out, name=f"{table.name}-melt")
        
    def summarize_table(self):
        """Summarize current table."""
        table = self.viewer.current_table
        out = summarize_table(
            df={"bind": table.data}, 
            methods={"choices": SUMMARY_CHOICES, "widget_type": "Select"}
        )
        if out is not None:
            self.viewer.add_table(out, name=f"{table.name}-summary")
    
    def query(self):
        """Filter table using a query."""
        table = self.viewer.current_table
        out = query(df={"bind": table.data})
        if out is not None:
            self.viewer.add_table(out, name=f"{table.name}-query")


@dialog_factory
def summarize_table(df: TableData, methods: List[str]):
    return df.agg(methods)

@dialog_factory
def groupby(df: TableData, by: List[str]):
    return df.groupby(by)

@dialog_factory
def hconcat(viewer, names: List[str]):
    dfs = [viewer.tables[name].data for name in names]
    return pd.concat(dfs, axis=0)

@dialog_factory
def vconcat(viewer, names: List[str]):
    dfs = [viewer.tables[name].data for name in names]
    return pd.concat(dfs, axis=1)

@dialog_factory
def pivot(df: TableData, index: str, columns: str, values: str):
    return df.pivot(index=index, columns=columns, values=values)

@dialog_factory
def melt(df: TableData, id_vars: List[str]):
    return pd.melt(df, id_vars)

@dialog_factory
def query(df: TableData, expr: str):
    return df.query(expr)



from __future__ import annotations
from typing import TYPE_CHECKING
import weakref
from qtpy import QtWidgets as QtW, QtGui, QtCore
from qtpy.QtCore import Qt

if TYPE_CHECKING:
    import pandas as pd

class QTreeHeaderView(QtW.QHeaderView):
    def __init__(self, orientation: Qt.Orientation, parent: QtW.QWidget | None = None) -> None:
        super().__init__(orientation, parent)
    
    def paintSection(self, painter: QtGui.QPainter, rect: QtCore.QRect, logicalIndex: int) -> None:
        return super().paintSection(painter, rect, logicalIndex)

class QTableHeaderItem(QtW.QTableWidgetItem):
    def __init__(self, parent, item_parent=None):
        super().__init__(parent)
        self._item_parent = item_parent
        
    def insertChild(self):
        ...
    
    def children(self):
        ...
    
   
class HeaderModel(QtW.QAbstractItemView):
    
    ...
    
    # TODO: 
    # df0.columns.levels
    # Out[12]: FrozenList([['b', 'c', 'x'], ['mean', 'std', '']])

    # mi = df0.columns

    # mi.get_indexer
    # Out[14]: 
    # <bound method Index.get_indexer of MultiIndex([('b', 'mean'),
    #             ('b',  'std'),
    #             ('c', 'mean'),
    #             ('c',  'std'),
    #             ('x',     '')],
    #            )>

    # mi.get_level_values(0)
    # Out[15]: Index(['b', 'b', 'c', 'c', 'x'], dtype='object')

    # mi.get_level_values(1)
    # Out[16]: Index(['mean', 'std', 'mean', 'std', ''], dtype='object')

    # mi.nlevels
    # Out[17]: 2



from ._model_base import AbstractDataFrameModel, DataFrameModel
from ._table_base import (
    QBaseTable,
    QMutableTable,
    QMutableSimpleTable,
    _QTableViewEnhanced,
)

__all__ = [
    "AbstractDataFrameModel",
    "QBaseTable",
    "QMutableTable",
    "QMutableSimpleTable",
    "_QTableViewEnhanced",
]



from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Any, Callable, Generic, TYPE_CHECKING, TypeVar
from functools import partial
from psygnal import SignalGroup, Signal

from .keybindings import register_shortcut
from .filtering import FilterProperty

from ..types import SelectionRanges

if TYPE_CHECKING:
    import pandas as pd
    from .._qt import QTableLayer, QSpreadSheet, QTableGroupBy, QTableDisplay
    from .._qt._table import QBaseTable


class TableSignals(SignalGroup):
    """Signal group for a Table."""

    data = Signal(object)
    selections = Signal(object)
    zoom = Signal(float)
    renamed = Signal(str)


_QW = TypeVar("_QW", bound="QBaseTable")


class TableBase(ABC, Generic[_QW]):
    """The base class for a table layer."""

    _Default_Name = "None"

    def __init__(self, data, name=None, editable: bool = True):
        self._data = self._normalize_data(data)
        from .._qt._table import QMutableTable

        if name is None:
            name = self._Default_Name
        self.events = TableSignals()
        self._name = str(name)
        self._qwidget = self._create_backend(self._data)
        self._qwidget.connectSelectionChangedSignal(self.events.selections.emit)

        if isinstance(self._qwidget, QMutableTable):
            self._qwidget.setEditability(editable)
            self._qwidget.connectItemChangedSignal(self.events.data.emit)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}<{self.name!r}>"

    @abstractmethod
    def _create_backend(self, data: pd.DataFrame) -> _QW:
        """This function creates a backend widget."""

    @abstractmethod
    def _normalize_data(self, data):
        """Data normalization before setting a new data."""

    @property
    def data(self) -> pd.DataFrame:
        """Table data."""
        return self._qwidget.getDataFrame()

    @data.setter
    def data(self, value):
        self._data = self._normalize_data(value)
        self._qwidget.setDataFrame(self._data)

    @property
    def table_shape(self) -> tuple[int, int]:
        """Shape of table."""
        return self._qwidget.tableShape()

    @property
    def precision(self) -> int:
        """Precision (displayed digits) of table."""
        return self._qwidget.precision()

    @precision.setter
    def precision(self, value: int) -> None:
        self._qwidget.setPrecision(value)

    @property
    def zoom(self) -> float:
        """Zoom factor of table."""
        return self._qwidget.zoom()

    @zoom.setter
    def zoom(self, value: float):
        self._qwidget.setZoom(value)

    def _set_name(self, value: str):
        with self.events.renamed.blocked():
            self.name = value

    @property
    def name(self) -> str:
        """Table name."""
        return self._name

    @name.setter
    def name(self, value: str):
        self._name = str(value)
        self.events.renamed.emit(self._name)

    @property
    def editable(self) -> bool:
        """Editability of table."""
        return self._qwidget.editability()

    @editable.setter
    def editable(self, value: bool):
        self._qwidget.setEditability(value)

    @property
    def columns(self):
        return self._data.columns

    @property
    def index(self):
        return self._data.index

    @property
    def selections(self):
        """Get the SelectionRanges object of current table selection."""
        rngs = SelectionRanges(self._qwidget.selections())
        rngs._changed.connect(self._qwidget.setSelections)
        rngs.events.removed.connect(lambda i, value: self._qwidget.setSelections(rngs))
        return rngs

    @selections.setter
    def selections(self, value) -> None:
        self._qwidget.setSelections(value)

    def refresh(self) -> None:
        """Refresh the table view."""
        return self._qwidget.refresh()

    filter = FilterProperty()

    def bind_key(self, *seq) -> Callable[[TableBase], Any | None]:
        """Bind callback function to a key sequence."""

        def register(f):
            register_shortcut(seq, self._qwidget, partial(f, self))

        return register


class _DataFrameTableLayer(TableBase[_QW]):
    """Table layer for DataFrame."""

    def _normalize_data(self, data) -> pd.DataFrame:
        import pandas as pd

        if not isinstance(data, pd.DataFrame):
            data = pd.DataFrame(data)
        return data


class TableView(_DataFrameTableLayer["QTableLayer"]):
    _Default_Name = "table"

    def _create_backend(self, data: pd.DataFrame) -> QTableLayer:
        from .._qt import QTableLayer

        return QTableLayer(data=data)


class SpreadSheet(_DataFrameTableLayer["QSpreadSheet"]):
    _Default_Name = "sheet"

    def _create_backend(self, data: pd.DataFrame) -> QSpreadSheet:
        from .._qt import QSpreadSheet

        return QSpreadSheet(data=data)


class GroupBy(TableBase["QTableGroupBy"]):
    _Default_Name = "groupby"

    def _create_backend(self, data: pd.DataFrame) -> QTableGroupBy:
        from .._qt import QTableGroupBy

        return QTableGroupBy(data=data)

    def _normalize_data(self, data):
        from pandas.core.groupby.generic import DataFrameGroupBy

        if not isinstance(data, DataFrameGroupBy):
            raise TypeError("Cannot only add DataFrameGroupBy object.")
        return data

    @property
    def group(self):
        return self._qwidget.currentGroup()

    @group.setter
    def group(self, val) -> None:
        return self._qwidget.setCurrentGroup(val)


class TableDisplay(TableBase["QTableDisplay"]):
    _Default_Name = "display"

    def _create_backend(self, data: Callable[[], Any]) -> QTableDisplay:
        from .._qt import QTableDisplay

        return QTableDisplay(loader=data)

    def _normalize_data(self, data):
        if not callable(data):
            raise TypeError("Can only add callable object.")
        return data

    @property
    def interval(self) -> int:
        return self._qwidget._qtimer.interval()

    @interval.setter
    def interval(self, value: int) -> None:
        if not 10 <= value <= 10000:
            raise ValueError("Interval must be between 10 and 10000.")
        self._qwidget._qtimer.setInterval(value)

    @property
    def loader(self) -> Callable[[], Any]:
        return self._qwidget.loader()

    @loader.setter
    def loader(self, value: Callable[[], Any]) -> None:
        self._qwidget.setLoader(value)



from __future__ import annotations
from typing import Any, Hashable, Iterable, Sequence, TYPE_CHECKING
from pandas.core.groupby.generic import DataFrameGroupBy
from qtpy import QtWidgets as QtW
from qtpy.QtCore import Signal, Qt

from ._base import QBaseTable, _QTableViewEnhanced, DataFrameModel

if TYPE_CHECKING:
    import pandas as pd


class _LabeledComboBox(QtW.QWidget):
    currentIndexChanged = Signal(int)

    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        self._values = list[Any]
        _layout = QtW.QHBoxLayout()
        _layout.setAlignment(Qt.AlignmentFlag.AlignLeft)
        _layout.setContentsMargins(2, 2, 2, 2)
        self.setLayout(_layout)
        self._cbox = QtW.QComboBox()
        self._label = QtW.QLabel()
        _layout.addWidget(self._label)
        _layout.addWidget(self._cbox)

        self._cbox.currentIndexChanged.connect(self.currentIndexChanged.emit)

    def setLabel(self, text: str) -> None:
        """Set the text of label."""
        return self._label.setText(text)

    def setChoices(self, items: Iterable[Any]):
        """Set choices of combo box."""
        self._values = list(items)
        self._cbox.clear()
        return self._cbox.addItems(map(str, self._values))

    def currentIndex(self) -> int:
        """Get the current index of the choice."""
        return self._cbox.currentIndex()

    def setCurrentIndex(self, index: int) -> None:
        """Set current index."""
        return self._cbox.setCurrentIndex(index)

    def CurrentValue(self) -> Any:
        """Current value."""
        return self._values[self.currentIndex()]

    def setCurrentValue(self, value: Any) -> None:
        try:
            index = self._values.index(value)
        except ValueError:
            raise ValueError(f"{value} is not a valid choice.")
        return self.setCurrentIndex(index)


class QTableGroupBy(QBaseTable):
    _data_raw: DataFrameGroupBy

    @property
    def _qtable_view(self) -> _QTableViewEnhanced:
        return self._qtable_view_

    def createQTableView(self):
        self._qtable_view_ = _QTableViewEnhanced()
        self._group_key_cbox = _LabeledComboBox()
        self._group_map: dict[Hashable, Sequence[int]] = {}
        self._group_key_cbox.currentIndexChanged.connect(
            lambda e: self.setFilter(self._filter_slice)
        )

        _layout = QtW.QVBoxLayout()
        _layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(_layout)
        self.layout().addWidget(self._group_key_cbox)
        self.layout().addWidget(self._qtable_view_)

    def getDataFrame(self) -> DataFrameGroupBy:
        return self._data_raw

    def setDataFrame(self, data: DataFrameGroupBy) -> None:
        if not isinstance(data, DataFrameGroupBy):
            raise TypeError(f"Data must be DataFrameGroupBy, not {type(data)}")
        self._data_raw = data

        # set label
        keys = self._data_raw.keys
        if isinstance(keys, list):
            if len(keys) == 1:
                label = keys[0]
            else:
                label = tuple(keys)
        else:
            label = keys
        self._group_key_cbox.setLabel(f"{label} = ")

        self._group_map = self._data_raw.groups
        self._group_key_cbox.setChoices(self._group_map.keys())
        self.setFilter(None)
        self._qtable_view.viewport().update()
        return

    def createModel(self):
        model = DataFrameModel(self)
        self._qtable_view.setModel(model)
        return None

    def tableSlice(self) -> pd.DataFrame:
        df: pd.DataFrame = self._data_raw.obj
        sl = self._group_map[self._group_key_cbox.CurrentValue()]
        return df.iloc[sl, :]

    def currentGroup(self) -> Hashable:
        index = self._group_key_cbox.currentIndex()
        return self._group_key_cbox._values[index]

    def setCurrentGroup(self, group: Hashable) -> None:
        return self._group_key_cbox.setCurrentValue(group)
